import argparse
import json
import logging.config
import time


from parsagon.api import (
    delete_pipeline,
    add_examples_to_custom_function,
    get_pipeline,
    get_pipeline_by_id,
    get_pipelines,
    poll_extract,
)
from parsagon.assistant import assist
from parsagon.api import delete_pipeline, add_examples_to_custom_function, get_pipeline, get_pipelines, poll_extract
from parsagon.create import create_program
from parsagon.edit import edit_program
from parsagon.exceptions import ParsagonException
from parsagon.executor import Executor, custom_functions_to_descriptions
from parsagon.runs import run
from parsagon.settings import get_api_key, save_setting, configure_logging
from parsagon.print import ask, assistant_print, status, input


logger = logging.getLogger(__name__)


def get_args(argv):
    parser = argparse.ArgumentParser(
        prog="parsagon", description="Scrapes and interacts with web pages based on natural language.", add_help=False
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="run the task in verbose mode")
    subparsers = parser.add_subparsers()

    # Create
    parser_create = subparsers.add_parser("create", description="Creates a program.")
    parser_create.add_argument(
        "--headless",
        action="store_true",
        help="run the browser in headless mode",
    )
    parser_create.add_argument(
        "--undetected",
        action="store_true",
        help="run in undetected mode",
    )
    parser_create.set_defaults(func=create_cli)

    # Edit
    parser_edit = subparsers.add_parser("edit", description="Edits a program.")
    parser_edit.set_defaults(func=edit)
    parser_edit.add_argument(
        "program_name",
        type=str,
        help="the name of the program to update",
    )

    # Detail
    parser_detail = subparsers.add_parser(
        "detail",
        description="Outputs details of a created program.",
    )
    parser_detail.add_argument(
        "-p",
        "--program",
        dest="program_name",
        type=str,
        help="the name of the program",
    )
    parser_detail.set_defaults(func=detail)

    # Update
    parser_update = subparsers.add_parser(
        "update",
        description="Updates a created program.",
    )
    parser_update.add_argument(
        "program_name",
        type=str,
        help="the name of the program to update",
    )
    parser_update.add_argument(
        "--variables",
        type=json.loads,
        default="{}",
        help="a JSON object mapping variables to values",
    )
    parser_update.add_argument(
        "--headless",
        action="store_true",
        help="run the browser in headless mode",
    )
    parser_update.add_argument(
        "--undetected",
        action="store_true",
        help="run in undetected mode",
    )
    parser_update.add_argument(
        "--infer",
        action="store_true",
        help="let Parsagon infer all elements to be scraped",
    )
    parser_update.add_argument(
        "--replace",
        action="store_true",
        help="remove old example data while updating the program",
    )
    parser_update.set_defaults(func=update)

    # Run
    parser_run = subparsers.add_parser(
        "run",
        description="Runs a created program.",
    )
    parser_run.add_argument(
        "program_name",
        type=str,
        help="the name of the program to run",
    )
    parser_run.add_argument(
        "--variables",
        type=json.loads,
        default="{}",
        help="a JSON object mapping variables to values",
    )
    parser_run.add_argument(
        "--headless",
        action="store_true",
        help="run the browser in headless mode",
    )
    parser_run.add_argument(
        "--remote",
        action="store_true",
        help="run the program in the cloud",
    )
    parser_run.add_argument(
        "--output_log",
        action="store_true",
        help="output log data from the run",
    )
    parser_run.add_argument(
        "--output_file",
        type=str,
        help="write the data to the given file path",
    )
    parser_run.add_argument(
        "--use_cdp",
        action="store_true",
        help="enable CDP (only in undetected mode)",
    )
    parser_run.add_argument(
        "--undetected",
        action="store_true",
        help="run in undetected mode",
    )
    parser_run.add_argument(
        "--optimize",
        action="store_true",
        help="run in optimized mode",
    )
    parser_run.add_argument(
        "--use_proxy",
        action="store_true",
        help="run with proxy",
    )
    parser_run.set_defaults(func=run)

    # Delete
    parser_delete = subparsers.add_parser(
        "delete",
        description="Deletes a program.",
    )
    parser_delete.add_argument(
        "program_name",
        type=str,
        help="the name of the program to run",
    )
    parser_delete.add_argument(
        "-y", "--yes", dest="confirm_with_user", action="store_false", help="auto-confirm option"
    )
    parser_delete.set_defaults(func=delete)

    # Setup
    parser_setup = subparsers.add_parser(
        "setup",
        description="Interactively sets up Parsagon with an API key.",
    )
    parser_setup.set_defaults(func=setup)

    # Help
    parser_help = subparsers.add_parser(
        "help",
        description="Shows help.",
    )
    parser_help.set_defaults(func=help, parser=parser)

    args = parser.parse_args(argv)
    kwargs = vars(args)
    return kwargs, parser


def main(argv=None):
    kwargs, parser = get_args(argv)
    func = kwargs.pop("func", None)
    if func is None:
        func = assist
    verbose = kwargs["verbose"]
    configure_logging(verbose)

    try:
        return func(**kwargs)
    except ParsagonException as e:
        error_message = "Error:\n" + e.to_string(verbose)
        logger.error(error_message)


def create_cli(headless=False, undetected=False, verbose=False):
    task = ask("Enter a detailed scraping task")
    create_program(task, headless=headless, undetected=undetected)


def edit(program_name, variables={}, verbose=False):
    task = ask("Describe how you want to edit the program")
    edit_program(task, program_name)


def update(
    program_name=None,
    program_id=None,
    variables={},
    headless=False,
    undetected=False,
    infer=False,
    replace=False,
    verbose=False,
):
    configure_logging(verbose)

    if program_id:
        pipeline = get_pipeline_by_id(program_id)
        program_name = pipeline["name"]
    else:
        pipeline = get_pipeline(program_name)
    abridged_program = pipeline["abridged_sketch"]
    # Make the program runnable
    variables_str = ", ".join(f"{k}={repr(v)}" for k, v in variables.items())
    abridged_program += f"\n\noutput = func({variables_str})\n"

    # Execute the abridged program to gather examples
    executor = Executor(pipeline["description"], headless=headless, infer=infer, use_uc=undetected)
    executor.execute(abridged_program)

    while True:
        program_name_input = input(
            f'Type "{program_name}" to update this program, or press enter without typing a name to CANCEL: '
        )
        if not program_name_input:
            logger.info("Canceled update.")
            return
        if program_name_input == program_name:
            break

    pipeline_id = pipeline["id"]
    try:
        for call_id, custom_function in executor.custom_functions.items():
            debug_suffix = f" ({custom_function.name})"
            description = custom_functions_to_descriptions.get(custom_function.name)
            description = " to " + description if description else ""
            if verbose:
                description += debug_suffix
            logger.info(f"  Saving function{description}...")
            add_examples_to_custom_function(pipeline_id, call_id, custom_function, replace)
        logger.info(f"Saved.")
    except Exception as e:
        logger.error(f"An error occurred while saving the program. The program was not updated.")


def detail(program_name=None, verbose=False):
    if program_name:
        data = [get_pipeline(program_name)]
    else:
        data = get_pipelines()
    for pipeline in data:
        assistant_print(
            f"Program: {pipeline['name']}\nDescription: {pipeline['description']}\nVariables: {pipeline['variables']}\n"
        )


def delete(program_name, verbose=False, confirm_with_user=False):
    if (
        confirm_with_user
        and input(f"Are you sure you want to delete program with name {program_name}? (y/N) ").lower().strip() != "y"
    ):
        logger.error("Cancelled operation.")
        return
    logger.info("Preparing to delete program %s", program_name)
    pipeline_id = get_pipeline(program_name)["id"]
    logger.info("Deleting program...")
    delete_pipeline(pipeline_id)
    logger.info("Done.")


def setup(verbose=False):
    try:
        old_api_key = get_api_key()
    except ParsagonException:
        old_api_key = None
    try:
        save_setting("api_key", None)
        get_api_key(interactive=True)
    except KeyboardInterrupt:
        save_setting("api_key", old_api_key)
        logger.error("\nCancelled operation.")
        return
    logger.info("Setup complete.")


def help(parser, verbose):
    parser.print_help()


def _get_data(url, page_type, timeout):
    start_time = time.time()
    with status("Extracting data..."):
        while time.time() - start_time <= timeout:
            result = poll_extract(url, page_type)
            if result["done"]:
                return result["result"]
            time.sleep(15)
    logger.info("No data found")
    return None


def get_product(url, timeout=300):
    return _get_data(url, "PRODUCT_DETAIL", timeout)


def get_review_article(url, timeout=300):
    return _get_data(url, "REVIEW_ARTICLE_DETAIL", timeout)


def get_article_list(url, timeout=300):
    return _get_data(url, "ARTICLE_LIST", timeout)



import base64
from collections import defaultdict
import copy
import dateutil.parser
import glob
import json
import logging
import os
from pathlib import Path
import psutil
import time
from urllib.parse import urljoin

import html2text
import httpx
from lxml import etree
import lxml.html
from lxml.html.clean import Cleaner
import pymupdf4llm
from pypdf import PdfReader
from pyvirtualdisplay import Display
import undetected_chromedriver as uc
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.select import Select
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.driver_cache import DriverCacheManager

from parsagon import settings
from parsagon.api import (
    get_interaction_element_id,
    get_schema_fields,
    get_cleaned_data,
    scrape_page,
    get_str_about_data,
    get_bool_about_data,
    get_json_about_data,
    get_serp_data,
)
from parsagon.custom_function import CustomFunction
from parsagon.exceptions import ParsagonException
from parsagon.print import browser_print, input, Progress, ask, confirm
from parsagon.settings import get_resource_path

# A dictionary of custom function names to their descriptions for the user
custom_functions_to_descriptions = {
    "scrape_data": "scrape data from the page",
    "goto": "go to a page",
    "click_elem": "click an element",
    "fill_input": "fill an input",
    "select_option": "select an option",
}


# A dict of schema types to element types
ELEMENT_TYPES = {
    "str": "TEXT",
    "num": "TEXT",
    "link": "URL",
    "image": "IMAGE",
    "html": "HTML",
    "element": "ACTION",
    "textarea": "TEXT",
    "text_markdown": "TEXT",
    "markdown": "TEXT",
    "elem_id": "ACTION",
}


class Executor:
    """
    Executes code produced by GPT with the proper context.  Records custom_function usage along the way.
    """

    def __init__(
        self,
        task,
        headless=False,
        infer=False,
        use_uc=False,
        driver_path=None,
        options=[],
        page_load_timeout=None,
        script_timeout=None,
        function_bank={},
    ):
        self.task = task
        self.headless = headless
        self.function_bank = function_bank
        if self.headless:
            self.display = Display(visible=False, size=(1280, 1050)).start()
        if driver_path:
            cache_manager = DriverCacheManager(root_dir=driver_path)
            driver_executable_path = ChromeDriverManager(cache_manager=cache_manager).install()
        else:
            driver_executable_path = ChromeDriverManager().install()
        if use_uc:
            chrome_options = uc.ChromeOptions()
            chrome_options.add_argument("--start-maximized")
            for option in options:
                chrome_options.add_argument(option)
            chrome_options.add_experimental_option(
                "prefs",
                {
                    "download.default_directory": os.getcwd(),
                    "download.prompt_for_download": False,
                    "download.directory_upgrade": True,
                    "plugins.always_open_pdf_externally": True,
                },
            )
            self.driver = uc.Chrome(driver_executable_path=driver_executable_path, options=chrome_options)
        else:
            chrome_options = webdriver.ChromeOptions()
            chrome_options.add_argument("--start-maximized")
            for option in options:
                chrome_options.add_argument(option)
            chrome_options.add_experimental_option(
                "prefs",
                {
                    "download.default_directory": os.getcwd(),
                    "download.prompt_for_download": False,
                    "download.directory_upgrade": True,
                    "plugins.always_open_pdf_externally": True,
                },
            )
            self.driver = webdriver.Chrome(service=ChromeService(driver_executable_path), options=chrome_options)
        if page_load_timeout:
            self.driver.set_page_load_timeout(page_load_timeout)
        if script_timeout:
            self.driver.set_script_timeout(script_timeout)
        self.max_elem_ids = defaultdict(int)
        self.execution_context = {
            "custom_assert": self.custom_assert,
            "goto": self.goto,
            "goto_lite": self.goto_lite,
            "close_window": self.close_window,
            "click_elem": self.click_elem,
            "click_elem_by_id": self.click_elem_by_id,
            "click_next_page": self.click_next_page,
            "fill_input": self.fill_input,
            "fill_input_by_id": self.fill_input_by_id,
            "select_option": self.select_option,
            "select_option_by_id": self.select_option_by_id,
            "scroll": self.scroll,
            "press_key": self.press_key,
            "join_text": self.join_text,
            "wait": self.wait,
            "get_inner_text": self.get_inner_text,
            "get_inner_text_by_id": self.get_inner_text_by_id,
            "scrape_data": self.scrape_data,
            "get_str_about_data": get_str_about_data,
            "get_bool_about_data": get_bool_about_data,
            "get_json_about_data": get_json_about_data,
            "get_serp_data": get_serp_data,
            "get_pdf_text": self.get_pdf_text,
            "str_to_iso8601": self.str_to_iso8601,
            "get_network_requests": self.get_network_requests,
        }
        self.custom_functions = {}
        self.infer = infer

        highlights_path = get_resource_path() / "highlights.js"
        with highlights_path.open() as f:
            self.highlights_script = f.read()

    def add_custom_function(self, call_id, custom_function):
        if call_id in self.custom_functions:
            self.custom_functions[call_id].examples.extend(custom_function.examples)
        else:
            self.custom_functions[call_id] = custom_function

    def exec_custom_function(self, name, call_id, args_dict):
        code = self.function_bank[call_id]
        args_dict["call_id"] = call_id
        args_str = ", ".join(f"{k}={repr(v)}" for k, v in args_dict.items())
        code = f"import time\nimport lxml.html\nfrom urllib.parse import urljoin\nfrom selenium.webdriver.common.by import By\n{code}\noutput = {name}{call_id}({args_str})"
        context = {
            "driver": self.driver,
            "PARSAGON_API_KEY": settings.get_api_key(),
            "parsagon_elem_data": {"idx": self.max_elem_ids[self.driver.current_window_handle]},
            "parsagon_log": [],
            "parsagon_warnings": [],
        }
        exec(code, context, context)
        return context["output"]

    def inject_highlights_script(self):
        self.driver.execute_script(self.highlights_script)

    def highlights_setup(self, field_type, max_examples="null"):
        self.driver.execute_script(f"window.currentFieldType = '{field_type}'; window.maxExamples = {max_examples};")

    def highlights_cleanup(self):
        self.driver.execute_script(f"window.currentFieldType = null; window.maxExamples = null; window.clearCSS();")

    def get_selected_node_ids(self, css_selector=None, xpath_selector=None):
        if css_selector:
            return [
                elem.get_attribute("data-psgn-id") for elem in self.driver.find_elements(By.CSS_SELECTOR, css_selector)
            ]
        elif xpath_selector:
            return [elem.get_attribute("data-psgn-id") for elem in self.driver.find_elements(By.XPATH, xpath_selector)]
        else:
            return [
                elem.get_attribute("data-psgn-id")
                for elem in self.driver.find_elements(By.CLASS_NAME, "parsagon-io-example-stored")
            ]

    def get_selected_node_and_descendant_ids(self):
        return self.driver.execute_script(
            "return Array.from(document.getElementsByClassName('parsagon-io-example-stored')).map((elem) => [elem, ...elem.querySelectorAll('*')]).flat().map((elem) => elem.getAttribute('data-psgn-id'))"
        )

    def mark_html(self):
        """
        Adds node IDs to elements on the current page that don't already have IDs.
        """
        max_elem_id = self.max_elem_ids[self.driver.current_window_handle]
        max_elem_id = self.driver.execute_script(
            f"let elemIdx = {max_elem_id}; "
            + "for (const node of document.all) { if (node.hasAttribute('data-psgn-id')) { continue; } node.setAttribute('data-psgn-id', elemIdx); elemIdx++; } return elemIdx;"
        )
        self.max_elem_ids[self.driver.current_window_handle] = max_elem_id
        self.driver.execute_script(
            "for (const image of document.images) { image.setAttribute('data-psgn-width', image.parentElement.offsetWidth ?? -1); image.setAttribute('data-psgn-height', image.parentElement.offsetHeight ?? -1); }"
        )

    def _get_cleaned_lxml_root(self):
        parser = lxml.html.HTMLParser(remove_comments=True, remove_pis=True)
        root = lxml.html.fromstring(self.driver.page_source.replace("&nbsp;", " "), parser=parser)

        # make links absolute
        root.make_links_absolute(self.driver.current_url)
        for elem in root.xpath("//img[@srcset]"):
            srcset_list = []
            for s in elem.get("srcset").split(","):
                parts = s.strip().split()
                if not parts:
                    continue
                parts[0] = urljoin(self.driver.current_url, parts[0])
                srcset_list.append(" ".join(parts))
            elem.set("srcset", ", ".join(srcset_list))

        # remove unnecessary and bulky elements
        for elem in root.iterfind(".//script"):
            elem.text = ""
        for elem in root.iterfind(".//noscript"):
            elem.text = ""
        for elem in root.iterfind(".//style"):
            elem.text = ""
        return root

    def get_scrape_html(self):
        """
        Returns cleaned html from the driver with script, noscript, and style elements removed, designed to preserve scrapable data.
        """
        root = self._get_cleaned_lxml_root()
        return lxml.html.tostring(root).decode()

    def get_visible_html(self):
        """
        Returns cleaned html from the driver, hiding all elements that are not visible.
        Script, noscript, style, and head elements are removed.
        All elements must have node IDs added as data attributes.
        """

        driver = self.driver
        assert "data-psgn-id" in driver.page_source
        root = self._get_cleaned_lxml_root()

        # Remove head elements
        for elem in root.iterfind(".//head"):
            elem.getparent().remove(elem)

        # Remove invisible elements
        id_to_elem = {}
        for elem in root.iter():
            elem_id = elem.get("data-psgn-id")
            if elem_id:
                id_to_elem[elem_id] = elem
        visible_elem_ids = set(
            driver.execute_script(
                "return Array.from(document.getElementsByTagName('*')).filter((elem) => { const style = getComputedStyle(elem); const rect = elem.getBoundingClientRect(); return style.opacity > 0.1 && style.display !== 'none' && style.visibility === 'visible' && elem.offsetWidth && elem.offsetHeight && elem.getClientRects().length && rect.right + window.scrollX >= 0 && rect.bottom + window.scrollY >= 0 }).map((elem) => elem.getAttribute('data-psgn-id'))"
            )
        )
        max_elem_id = self.max_elem_ids[self.driver.current_window_handle]
        with Progress() as progress:
            for elem_id in progress.track(range(max_elem_id), description="[green]Analyzing page"):
                elem_id = str(elem_id)
                is_visible = elem_id in visible_elem_ids
                if not is_visible:
                    try:
                        lxml_elem = root.xpath(f'//*[@data-psgn-id="{elem_id}"]')[0]
                        contains_visible = False
                        for descendant in lxml_elem.iterdescendants():
                            if descendant.get("data-psgn-id") in visible_elem_ids:
                                contains_visible = True
                                break
                        if contains_visible:
                            continue
                        parent = lxml_elem.getparent()
                        if parent is not None:
                            parent.remove(lxml_elem)
                    except IndexError:
                        continue
        etree.indent(root)
        return lxml.html.tostring(root).decode()

    def get_elem(self, description, elem_type):
        if self.infer:
            return self.get_elem_by_description(description, elem_type)
        self.highlights_setup("ACTION", max_examples=1)
        user_input = input(
            f'Click the element referred to by "{description}". Hit ENTER to confirm your selection, or type "N/A" if the element does not exist: '
        )
        self.mark_html()
        selected_node_ids = self.get_selected_node_ids()
        while (
            user_input != "N/A"
            and not selected_node_ids
            and not user_input.startswith("XPATH:")
            and not user_input.startswith("CSS:")
        ):
            user_input = input('Please click an element or type "N/A": ')
            selected_node_ids = self.get_selected_node_ids()
        self.highlights_cleanup()
        if user_input == "N/A":
            return None, None, None, None

        css_selector = None
        xpath_selector = None
        if user_input.startswith("CSS:"):
            css_selector = user_input[4:].strip()
            selected_node_ids = self.get_selected_node_ids(css_selector=css_selector)
        elif user_input.startswith("XPATH:"):
            xpath_selector = user_input[6:].strip()
            selected_node_ids = self.get_selected_node_ids(xpath_selector=xpath_selector)

        elem_id = selected_node_ids[0] if selected_node_ids else None
        elem = self._id_to_elem(elem_id) if selected_node_ids else None
        return elem, elem_id, css_selector, xpath_selector

    def get_elem_by_description(self, description, elem_type):
        browser_print(f'Looking for {elem_type.lower()}: "{description}"')
        self.mark_html()
        visible_html = self.get_visible_html()
        elem_id = get_interaction_element_id(visible_html, description, self.task)
        if elem_id is None:
            raise ParsagonException(
                f'Could not find an element matching "{description}". Perhaps try rephrasing your prompt.'
            )
        elem = self._id_to_elem(elem_id)
        log_suffix = f' with text "{elem.text}"' if elem.text else ""
        browser_print(f"Found element" + log_suffix)
        return elem, elem_id, None, None

    def _id_to_elem(self, elem_id):
        """
        Gets a selenium element by Parsagon ID (psgn-id).
        """
        assert elem_id is not None
        result = self.driver.find_element(By.XPATH, f'//*[@data-psgn-id="{elem_id}"]')
        return result

    def custom_assert(self, v):
        assert v, "Web page interaction failed."

    def goto(self, url, window_id=None):
        if window_id in self.driver.window_handles:
            self.driver.switch_to.window(window_id)
        else:
            self.driver.switch_to.new_window("tab")

        # Go to website
        browser_print(f"Going to {url}")
        self.driver.get(url)

        # Wait for website to load
        time.sleep(2)
        self.mark_html()
        self.inject_highlights_script()

        return self.driver.current_window_handle

    def goto_lite(self, url, window_id=None):
        if window_id in self.driver.window_handles:
            self.driver.switch_to.window(window_id)
        else:
            self.driver.switch_to.new_window("tab")
        # This is usually called in programs that use a proxy, but executor does not use proxies
        with httpx.Client(verify=False) as client:
            r = client.get(url, timeout=60)
        cleaner = Cleaner()
        cleaner.javascript = True
        html = lxml.html.tostring(cleaner.clean_html(lxml.html.fromstring(r.text)))
        self.driver.get("data:text/html;base64," + base64.b64encode(html).decode())
        time.sleep(1)
        return self.driver.current_window_handle

    def close_window(self, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        self.driver.close()
        self.driver.switch_to.window(self.driver.window_handles[-1])

    def _click_elem(self, elem, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)

        try:
            self.driver.execute_script("arguments[0].click();", elem)
            browser_print("Clicked element")
            time.sleep(2)
        except Exception as e:
            return False
        self.mark_html()
        self.inject_highlights_script()
        return True

    def click_elem(self, description, window_id, call_id):
        """
        Clicks a button using its description.
        """
        if call_id in self.function_bank:
            edit = confirm(f'Now clicking the element referred to by "{description}". Do you want to edit this step?')
            if not edit:
                return self.exec_custom_function(
                    "click_elem", call_id, {"description": description, "window_id": window_id}
                )
        elem, elem_id, css_selector, xpath_selector = self.get_elem(description, "BUTTON")
        html = self.get_scrape_html()
        success = self._click_elem(elem, window_id) if elem else False
        custom_function = CustomFunction(
            "click_elem",
            arguments={},
            examples=[
                {
                    "html": html,
                    "url": self.driver.current_url,
                    "elem_id": elem_id,
                    "css_selector": css_selector,
                    "xpath_selector": xpath_selector,
                }
            ],
        )
        self.add_custom_function(call_id, custom_function)
        return success

    def click_elem_by_id(self, elem_id, window_id):
        elem = self._id_to_elem(elem_id)
        return self._click_elem(elem, window_id)

    def click_next_page(self, description, window_id, call_id, wait=1):
        elem, elem_id, css_selector, xpath_selector = self.get_elem(description, "BUTTON")
        html = self.get_scrape_html()
        prev_html = self.driver.page_source
        success = self._click_elem(elem, window_id) if elem else False
        time.sleep(wait)
        custom_function = CustomFunction(
            "click_next_page",
            arguments={},
            examples=[
                {
                    "html": html,
                    "url": self.driver.current_url,
                    "elem_id": elem_id,
                    "xpath_selector": xpath_selector,
                }
            ],
        )
        self.add_custom_function(call_id, custom_function)
        return success and prev_html != self.driver.page_source

    def _select_option(self, elem, option, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)

        for i in range(3):
            try:
                select_obj = Select(elem)
                select_obj.select_by_visible_text(option)
                browser_print(f'Selected option "{option}"')
                time.sleep(2)
                break
            except:
                time.sleep(2)
        else:
            return False
        self.mark_html()
        self.inject_highlights_script()
        return True

    def select_option(self, description, option, window_id, call_id):
        """
        Selects an option by name from a dropdown using its description.
        """
        elem, elem_id, css_selector, xpath_selector = self.get_elem(description, "SELECT")
        html = self.get_scrape_html()
        success = self._select_option(elem, option, window_id) if elem else False
        custom_function = CustomFunction(
            "select_option",
            arguments={
                "option": option,
            },
            examples=[
                {
                    "html": html,
                    "url": self.driver.current_url,
                    "elem_id": elem_id,
                    "css_selector": css_selector,
                    "xpath_selector": xpath_selector,
                }
            ],
        )
        self.add_custom_function(call_id, custom_function)
        return success

    def select_option_by_id(self, elem_id, option, window_id):
        elem = self._id_to_elem(elem_id)
        return self._select_option(elem, option, window_id)

    def _fill_input(self, elem, text, enter, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)

        for i in range(3):
            try:
                elem.clear()
                elem.send_keys(text)
                browser_print(f'Typed "{text}" into element')
                if enter:
                    elem.send_keys(Keys.RETURN)
                time.sleep(2)
                break
            except:
                time.sleep(2)
        else:
            return False
        self.mark_html()
        self.inject_highlights_script()
        return True

    def fill_input(self, description, text, enter, window_id, call_id):
        """
        Fills an input text field, then presses an optional end key using its description.
        """
        elem, elem_id, css_selector, xpath_selector = self.get_elem(description, "INPUT")
        html = self.get_scrape_html()
        success = self._fill_input(elem, text, enter, window_id) if elem else False
        custom_function = CustomFunction(
            "fill_input",
            arguments={
                "text": text,
                "enter": enter,
            },
            examples=[
                {
                    "html": html,
                    "url": self.driver.current_url,
                    "elem_id": elem_id,
                    "css_selector": css_selector,
                    "xpath_selector": xpath_selector,
                }
            ],
        )
        self.add_custom_function(call_id, custom_function)
        return success

    def fill_input_by_id(self, elem_id, text, enter, window_id):
        elem = self._id_to_elem(elem_id)
        return self._fill_input(elem, text, enter, window_id)

    def scroll(self, x, y, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        browser_print(f"Scrolling {x * 100}% to the left and {y * 100}% down")
        self.driver.execute_script(
            f"window.scrollTo({{top: document.documentElement.scrollHeight * {y}, left: document.documentElement.scrollWidth * {x}, behavior: 'smooth'}});"
        )
        time.sleep(1)
        self.mark_html()

    def press_key(self, key, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        browser_print(f"Pressing {key}")
        ActionChains(self.driver).send_keys(getattr(Keys, key)).perform()
        time.sleep(1)

    def join_text(strings):
        return "\\n\\n".join(strings)

    def wait(self, seconds):
        browser_print(f"Waiting {seconds} seconds...")
        time.sleep(seconds)
        self.mark_html()
        self.inject_highlights_script()

    def get_inner_text(self, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        return self.driver.execute_script("return document.body.innerText;")

    def get_inner_text_by_id(self, elem_id, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        elem = self._id_to_elem(elem_id)
        return elem.get_attribute("innerText")

    def get_inner_markdown(self, window_id):
        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)
        html = self.driver.execute_script("return document.body.innerHTML;")
        return html2text.html2text(html)

    def scrape_data(self, schema, window_id, call_id):
        """
        Scrapes data from the current page.
        """
        if call_id in self.function_bank:
            edit = confirm(f"Now scraping data in the format {schema}. Do you want to edit this step?")
            if not edit:
                return self.exec_custom_function("scrape_data", call_id, {"schema": schema, "window_id": window_id})

        if self.driver.current_window_handle != window_id:
            self.driver.switch_to.window(window_id)

        browser_print(f"Now collecting data in the format {schema}")

        self.mark_html()
        html = self.get_scrape_html()
        nodes = {}
        css_selectors = {}
        xpath_selectors = {}
        flagged_fields = {}
        if not self.infer:
            field_types = get_schema_fields(schema)
            for field, field_type in field_types.items():
                self.highlights_setup(ELEMENT_TYPES[field_type])
                field_repr = field.replace("dataset0|", "").replace("|", " / ")
                field_input = ask(
                    f"Click elements containing data for the field `{field_repr}`. Hit TAB to autocomplete or DELETE/BACKSPACE to clear selections. Hit ENTER when done"
                )
                if field_input.startswith("CSS:"):
                    css_selector = field_input[4:].strip()
                    nodes[field] = [[node_id] for node_id in self.get_selected_node_ids(css_selector=css_selector)]
                    css_selectors[field] = css_selector
                elif field_input.startswith("XPATH:"):
                    xpath_selector = field_input[6:].strip()
                    nodes[field] = [[node_id] for node_id in self.get_selected_node_ids(xpath_selector=xpath_selector)]
                    xpath_selectors[field] = xpath_selector
                else:
                    nodes[field] = [[node_id] for node_id in self.get_selected_node_ids()]
                self.highlights_cleanup()
            browser_print("Scraping data...")
            result = get_cleaned_data(html, schema, nodes)
            scraped_data = result["data"]
        else:
            browser_print("Scraping data...")
            result = scrape_page(self.get_visible_html(), schema, self.task, html)
            print(result)
            scraped_data = result["data"]
            nodes = result["nodes"]
            css_selectors = result["css"]
            xpath_selectors = result["xpath"]
            flagged_fields = result["flagged"]
            if not scraped_data or not nodes:
                raise ParsagonException(
                    f"Parsagon could not find any data on the page that would fit the format {schema}. You can try rephrasing your prompt, or you can run Parsagon in manual mode to click on elements you want to scrape."
                )
        browser_print(f"Scraped data:\n{scraped_data}\nFlagged fields: {[k for k, v in flagged_fields.items() if v]}\n")

        custom_function = CustomFunction(
            "scrape_data",
            arguments={
                "schema": schema,
            },
            examples=[
                {
                    "html": html,
                    "url": self.driver.current_url,
                    "nodes": nodes,
                    "css_selectors": css_selectors,
                    "xpath_selectors": xpath_selectors,
                    "flagged_fields": flagged_fields,
                    "scraped_data": copy.deepcopy(scraped_data),
                }
            ],
        )
        self.add_custom_function(call_id, custom_function)
        return scraped_data

    def get_pdf_text(self, url):
        window_id = self.goto(url)
        self.close_window(window_id)
        files = glob.glob("*")
        most_recent_file = max(files, key=os.path.getmtime)
        reader = PdfReader(most_recent_file)
        text = ""
        for page in reader.pages:
            text += page.extract_text(extraction_mode="layout", layout_mode_space_vertically=False)
            text += "\n"
        os.remove(most_recent_file)
        return text

    def get_pdf_md(self, url):
        window_id = self.goto(url)
        self.close_window(window_id)
        files = glob.glob("*")
        most_recent_file = max(files, key=os.path.getmtime)
        md_text = pymupdf4llm.to_markdown(most_recent_file)
        os.remove(most_recent_file)
        return md_text

    def str_to_iso8601(self, s):
        while s:
            try:
                return dateutil.parser.parse(s).isoformat()
            except dateutil.parser._parser.ParserError:
                s = " ".join(s.split()[:-1])
        return ""

    def get_network_requests(window_id, delay=30):
        requests = []

        def handle_request(event):
            requests.append(event)

        driver.switch_to.window(window_id)
        driver.add_cdp_listener("Network.requestWillBeSent", handle_request)
        wait(delay)
        return requests

    def execute(self, code):
        loc = {}
        try:
            exec(code, self.execution_context, loc)
            browser_print(f"Program finished and returned a value of:\n{loc['output']}\n")
            return loc["output"]
        finally:
            self.quit()

    def quit(self):
        self.driver.quit()
        for proc in psutil.process_iter():
            try:
                if proc.name() == "chromedriver":
                    proc.kill()
            except psutil.NoSuchProcess:
                continue
        if self.headless:
            self.display.stop()



import datetime
import json
import logging.config
import time
import traceback

import psutil

from parsagon.api import (
    create_pipeline_run,
    update_pipeline_run,
    get_pipeline,
    get_pipeline_code,
    get_run,
)
from parsagon.exceptions import ParsagonException, RunFailedException
from parsagon.settings import get_api_key
from parsagon.print import status, Progress

logger = logging.getLogger(__name__)


def run(
    program_name,
    program_id=None,
    variables={},
    headless=False,
    remote=False,
    output_log=False,
    output_file=None,
    use_cdp=False,
    undetected=False,
    optimize=False,
    use_proxy=False,
    global_variables={},
    verbose=False,
):
    """
    Executes pipeline code
    """
    if headless and remote:
        raise ParsagonException("Cannot run a program remotely in headless mode")

    if not isinstance(variables, dict):
        raise ParsagonException("Variables must be a dictionary")

    logger.info("Preparing to run program %s", program_name)
    pipeline_id = program_id or get_pipeline(program_name)["id"]

    if remote:
        result = create_pipeline_run(pipeline_id, variables, False)
        with status("Program running remotely..."):
            while True:
                run = get_run(result["id"])
                run_status = run["status"]

                if output_log and run_status in ("FINISHED", "ERROR"):
                    result = {k: v for k, v in run.items() if k in ("output", "status", "log", "warnings", "error")}
                    if output_file:
                        with open(output_file, "w") as f:
                            json.dump(result, f, indent=4)
                        return
                    else:
                        return result

                if run_status == "FINISHED":
                    if verbose:
                        logger.info(run["log"])
                        for warning in run["warnings"]:
                            logger.warning(warning)
                    logger.info("Program finished running.")
                    result = run["output"]
                    if output_file:
                        with open(output_file, "w") as f:
                            json.dump(result, f, indent=4)
                        return
                    else:
                        return result
                elif run_status == "ERROR":
                    raise ParsagonException(f"Program failed to run: {run['error']}")
                elif run_status == "CANCELED":
                    raise ParsagonException("Program execution was canceled")

                time.sleep(5)

    run = create_pipeline_run(pipeline_id, variables, True)
    code = get_pipeline_code(
        program_name, variables, headless, use_cdp, undetected, optimize, use_proxy, pipeline_id=program_id
    )["code"]
    start_time = datetime.datetime.now(datetime.timezone.utc).isoformat()
    run_data = {"start_time": start_time}

    logger.info("Running program...")
    globals_locals = {"PARSAGON_API_KEY": get_api_key(), **global_variables}
    try:
        exec(code, globals_locals, globals_locals)
        run_data["status"] = "FINISHED"
    except:
        run_data["status"] = "ERROR"
        run_data["error"] = str(traceback.format_exc())
        if not output_log:
            raise
    finally:
        end_time = datetime.datetime.now(datetime.timezone.utc).isoformat()
        run_data["end_time"] = end_time
        if "driver" in globals_locals:
            globals_locals["driver"].quit()
        if "display" in globals_locals:
            globals_locals["display"].stop()
        if "parsagon_log" in globals_locals:
            run_data["log"] = "\n".join(globals_locals["parsagon_log"])
            logger.info(run_data["log"])
        if "parsagon_warnings" in globals_locals:
            run_data["warnings"] = globals_locals["parsagon_warnings"]
        for proc in psutil.process_iter():
            try:
                if proc.name() == "chromedriver":
                    proc.kill()
            except psutil.NoSuchProcess:
                continue
        run = update_pipeline_run(run["id"], run_data)
    logger.info("Done.")

    if output_log:
        if "error" not in run_data:
            run["output"] = globals_locals["output"]
        result = {k: v for k, v in run.items() if k in ("output", "status", "log", "warnings", "error")}
    else:
        result = globals_locals["output"]

    if output_file:
        with open(output_file, "w") as f:
            json.dump(result, f, indent=4)
        return
    else:
        return result


def batch_runs(
    batch_name,
    program_name,
    runs,
    headless=False,
    ignore_errors=False,
    error_value=None,
    rerun_warnings=False,
    rerun_warning_types=[],
    rerun_errors=False,
    verbose=False,
):
    # Validate runs
    if not all(isinstance(run_, dict) for run_ in runs):
        raise ParsagonException("Runs must be a list of dictionaries")

    save_file = f"{batch_name}.json"
    try:
        with open(save_file) as f:
            outputs = json.load(f)
    except FileNotFoundError:
        outputs = []
    metadata_file = f"{batch_name}_metadata.json"
    try:
        with open(metadata_file) as f:
            metadata = json.load(f)
    except FileNotFoundError:
        metadata = []

    num_initial_results = len(outputs)
    error = None
    variables = None
    try:
        default_desc = f'Running program "{program_name}"'
        with Progress() as progress:
            task = progress.add_task(default_desc, total=len(runs))
            for i, variables in progress.track(enumerate(runs), task_id=task):
                if i < num_initial_results:
                    if rerun_errors and metadata[i]["status"] == "ERROR":
                        pass
                    elif rerun_warnings and metadata[i]["warnings"]:
                        if not rerun_warning_types or any(
                            warning["type"] in rerun_warning_types for warning in metadata[i]["warnings"]
                        ):
                            pass
                        else:
                            continue
                    else:
                        continue
                for j in range(3):
                    result = run(program_name, variables, headless, output_log=True)
                    if result["status"] != "ERROR":
                        output = result.pop("output")
                        if i < num_initial_results:
                            outputs[i] = output
                            metadata[i] = result
                        else:
                            outputs.append(output)
                            metadata.append(result)
                        break
                    else:
                        error = result["error"].strip().split("\n")[-1]
                        if j < 2:
                            progress.update(
                                task,
                                description=f"An error occurred: {error} - Waiting 60s before retrying (Attempt {j+2}/3)",
                            )
                            time.sleep(60)
                            progress.update(task, description=default_desc)
                            error = None
                            continue
                        else:
                            if ignore_errors:
                                error = None
                                if i < num_initial_results:
                                    outputs[i] = error_value
                                else:
                                    outputs.append(error_value)
                                break
                            else:
                                raise RunFailedException
    except RunFailedException:
        pass
    except Exception as e:
        error = repr(e)
    finally:
        if error:
            logger.error(
                f"Unresolvable error occurred on run with variables {variables}: {error} - Data has been saved to {save_file}. Rerun your command to resume."
            )
        with open(save_file, "w") as f:
            json.dump(outputs, f)
        with open(metadata_file, "w") as f:
            json.dump(metadata, f)
    num_warnings = 0
    num_runs_with_warnings = 0
    for m in metadata:
        if m["warnings"]:
            num_warnings += len(m["warnings"])
            num_runs_with_warnings += 1
    logger.info(
        f"\nSummary: {len(outputs)} runs made; {num_warnings} warnings encountered across {num_runs_with_warnings} runs. See {metadata_file} for logs.\n"
    )
    return None if error else outputs



from json import JSONDecodeError
import time
from urllib.parse import quote

import httpx

from parsagon import settings
from parsagon.exceptions import APIException, ProgramNotFoundException
from parsagon.settings import get_save_api_key_interactive

environment = "PANDAS_1.x"


class RaiseProgramNotFound:
    def __init__(self, program_name):
        self.program_name = program_name

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None and issubclass(exc_type, APIException):
            if exc_value.status_code == 404:
                raise ProgramNotFoundException(self.program_name)
        return False


def _request_to_exception(response):
    status_code = response.status_code
    if status_code == 500:
        raise APIException("A server error occurred. Please notify Parsagon.", status_code)
    if status_code in (502, 503, 504):
        raise APIException("Lost connection to server.", status_code)
    try:
        errors = response.json()
        if "detail" in errors and errors["detail"] == "Invalid token.":
            raise APIException("Invalid API key.", status_code)
        if "non_field_errors" in errors:
            raise APIException(errors["non_field_errors"], status_code)
        else:
            raise APIException(errors, status_code)
    except JSONDecodeError:
        raise APIException("Could not parse response.", status_code)


def _api_call(httpx_func, endpoint, output="json", **kwargs):
    """
    Calls `httpx_func` using the given endpoint. kwargs are passed to `httpx_func`

    Output mode can be "json" or "none".
    "json" returns the response as a deserialized JSON object.
    "none" returns None.

    If the request is not successful, an exception is thrown.
    """
    api_key = settings.get_api_key()
    api_endpoint = f"{settings.get_api_base()}/api{endpoint}"
    headers = {"Authorization": f"Token {api_key}"}
    r = httpx_func(api_endpoint, headers=headers, timeout=None, **kwargs)
    if not r.is_success:
        _request_to_exception(r)
    if output == "json":
        return r.json()
    elif output == "none":
        return
    else:
        raise Exception("Invalid output mode")


def get_program_sketches(description):
    """
    Gets a program sketches (full and abridged) from a description.
    :param description: Description in natural language that will be used to generate the scraping program.
    :return: A dict with keys "full", "abridged", and "pseudocode" for the respective program ASTs and pseudocode.
    """
    return _api_call(httpx.post, "/transformers/get-program-sketch/", json={"description": description})


def edit_program_sketch(sketch, description):
    return _api_call(
        httpx.post, "/transformers/edit-program-sketch/", json={"sketch": sketch, "description": description}
    )


def get_interaction_element_id(marked_html, description, task):
    """
    Gets a program sketches (full and abridged) from a description.
    :param marked_html: HTML with data-psgn-id attributes.
    :param description: A natural language description of the element.
    :return: The integer ID (data-psgn-id) of the element in the marked HTML.
    """
    result = _api_call(
        httpx.post,
        "/transformers/get-nav-elem/",
        json={"html": marked_html, "description": description, "task": task},
    )["id"]
    return result


def get_schema_fields(schema):
    """
    Gets fields and their types from a given schema
    :param schema:
    :return: A dict mapping fields to their types
    """
    return _api_call(httpx.post, "/transformers/get-schema-fields/", json={"schema": schema})


def get_cleaned_data(html, schema, nodes):
    """
    Gets cleaned data from the nodes to be scraped.
    :param html: HTML of the page to scrape
    :param schema: Schema of the data to scrape
    :param nodes: Nodes to scrape
    :return: Cleaned data
    """
    return _api_call(
        httpx.post, "/transformers/get-cleaned-data/", json={"html": html, "schema": schema, "nodes": nodes}
    )


def scrape_page(html, schema, task, full_html=None):
    """
    Scrapes data from the provided page HTML - data will be returned in the schema provided.
    :param html: HTML of the page to scrape.
    :param schema: Schema of the data to scrape
    :param task: The user's task
    :return: Scraped data
    """
    return _api_call(
        httpx.post,
        "/transformers/get-custom-data/",
        json={"html": html, "schema": schema, "task": task, "full_html": full_html},
    )


def poll_about_data(url, json):
    for _ in range(20):
        result = _api_call(httpx.post, url, json=json)
        if result["done"]:
            return result["result"]
        time.sleep(5)
    raise APIException("Request timed out")


def get_str_about_data(data, question, model=None):
    """
    Asks GPT a question about the given data.
    :param data: the data to give GPT
    :param question: the question to ask about the data
    """
    return poll_about_data("/transformers/get-str-about-data/", {"data": data, "question": question, "model": model})


def get_bool_about_data(data, question, model=None):
    """
    Asks GPT a question about the given data.
    :param data: the data to give GPT
    :param question: the question to ask about the data
    """
    return poll_about_data("/transformers/get-bool-about-data/", {"data": data, "question": question, "model": model})


def get_json_about_data(data, question, model=None):
    """
    Asks GPT a question about the given data.
    :param data: the data to give GPT
    :param question: the question to ask about the data
    """
    return poll_about_data("/transformers/get-json-about-data/", {"data": data, "question": question, "model": model})


def get_serp_data(url):
    """
    Get Google results
    :param url: the search url
    """
    return _api_call(httpx.post, "/transformers/get-serp-data/", json={"url": url})


def create_pipeline(name, description, program_sketch, pseudocode, secrets):
    return _api_call(
        httpx.post,
        "/pipelines/",
        json={
            "name": name,
            "description": description,
            "program_sketch": program_sketch,
            "pseudocode": pseudocode,
            "secrets": secrets,
        },
    )


def update_pipeline(pipeline_id, data):
    return _api_call(httpx.patch, f"/pipelines/{pipeline_id}/", json=data)


def delete_pipeline(pipeline_id):
    _api_call(httpx.delete, f"/pipelines/{pipeline_id}/", output="none")


def create_custom_function(pipeline_id, call_id, custom_function):
    _api_call(
        httpx.post,
        "/transformers/custom-function/",
        json={"pipeline": pipeline_id, "call_id": call_id, **custom_function.to_json()},
    )


def add_examples_to_custom_function(pipeline_id, call_id, custom_function, remove_old_examples):
    _api_call(
        httpx.post,
        "/transformers/custom-function/add-examples/",
        json={
            "pipeline": pipeline_id,
            "call_id": call_id,
            "remove_old_examples": remove_old_examples,
            **custom_function.to_json(),
        },
    )


def double_quote(string):
    """
    A bug in Django disallows URLs even with quoted slashes as in:
    /api/pipelines/name/stripe%2Fstuff/code/ HTTP/1.1" 405
    Therefore we must double quote
    """
    return quote(quote(string, safe=""), safe="")


def get_pipeline(pipeline_name):
    escaped_pipeline_name = double_quote(pipeline_name)
    with RaiseProgramNotFound(pipeline_name):
        return _api_call(
            httpx.get,
            f"/pipelines/name/{escaped_pipeline_name}/?double_escaped=True",
        )


def get_pipeline_by_id(pipeline_id):
    return _api_call(httpx.get, f"/pipelines/{pipeline_id}/")


def get_pipelines():
    return _api_call(httpx.get, f"/pipelines/")


def get_pipeline_code(pipeline_name, variables, headless, use_cdp, use_uc, optimize, use_proxy, pipeline_id=None):
    if pipeline_id:
        return _api_call(
            httpx.post,
            f"/pipelines/{pipeline_id}/code/",
            json={
                "variables": variables,
                "headless": headless,
                "use_cdp": use_cdp,
                "use_uc": use_uc,
                "optimize": optimize,
                "use_proxy": use_proxy,
            },
        )
    escaped_pipeline_name = double_quote(pipeline_name)
    with RaiseProgramNotFound(pipeline_name):
        return _api_call(
            httpx.post,
            f"/pipelines/name/{escaped_pipeline_name}/code/?double_escaped=True",
            json={
                "variables": variables,
                "headless": headless,
                "use_cdp": use_cdp,
                "use_uc": use_uc,
                "optimize": optimize,
                "use_proxy": use_proxy,
            },
        )


def get_pipeline_code_parts(pipeline_id):
    return _api_call(httpx.get, f"/pipelines/{pipeline_id}/code-parts/")


def create_pipeline_run(pipeline_id, variables, is_local):
    return _api_call(
        httpx.post,
        f"/pipelines/{pipeline_id}/runs/",
        json={"variables": variables, "is_local": is_local},
    )


def update_pipeline_run(run_id, data):
    return _api_call(
        httpx.patch,
        f"/pipelines/runs/{run_id}/",
        json=data,
    )


def get_run(run_id):
    """
    Gets details about a run
    """
    return _api_call(
        httpx.get,
        f"/pipelines/runs/{run_id}/",
    )


def schedule(program_name, interval, runs=[{}]):
    pipeline_id = get_pipeline(program_name)["id"]
    return _api_call(
        httpx.post,
        f"/pipelines/schedules/",
        json={"pipeline": pipeline_id, "interval": interval, "variables_list": runs},
    )


def delete_schedule(program_name):
    pipeline_id = get_pipeline(program_name)["id"]
    return _api_call(
        httpx.delete,
        f"/pipelines/{pipeline_id}/schedules/",
    )


def send_assistant_message(message, thread_id=None):
    return _api_call(
        httpx.post, "/transformers/send-assistant-message/", json={"message": message, "thread_id": thread_id}
    )


def send_assistant_function_outputs(outputs, thread_id, run_id):
    return _api_call(
        httpx.post,
        "/transformers/send-assistant-function-outputs/",
        json={"outputs": outputs, "thread_id": thread_id, "run_id": run_id},
    )


def poll_extract(url, page_type):
    return _api_call(httpx.post, "/extract/", json={"url": url, "page_type": page_type})



from parsagon.main import update, detail, delete, get_product, get_review_article, get_article_list
from parsagon.create import create_program as create
from parsagon.runs import run, batch_runs
from parsagon.api import get_str_about_data, get_bool_about_data, get_json_about_data, get_pipeline
from parsagon.assistant import get_page_html, get_page_text
from parsagon.executor import Executor
from parsagon.settings import save_setting, save_settings



import pytest

from parsagon.main import main


def test_headless_remote_run_invalid(mocker, debug_logs):
    """
    Tests that we are unable to run a program in headless mode when the environment is remote, and that this is logged to the user.
    """
    main(["run", "test_program", "--headless", "--remote"])
    debug_logs_lower = debug_logs.text.lower()
    assert "error" in debug_logs_lower
    assert "headless" in debug_logs_lower
    assert "remote" in debug_logs_lower



import os
from datetime import datetime
from threading import Condition

from PyQt6.QtCore import Qt
from PyQt6.QtCore import pyqtSlot, QSize, QTimer
from PyQt6.QtGui import (
    QMovie,
    QKeyEvent,
    QFont,
    QPixmap,
    QIcon,
    QTextOption,
    QAction,
)
from PyQt6.QtWidgets import (
    QMainWindow,
    QVBoxLayout,
    QPushButton,
    QWidget,
    QHBoxLayout,
    QTextEdit,
    QLabel,
    QProgressBar,
    QScrollArea,
    QSpacerItem,
    QSizePolicy,
    QFrame,
    QMessageBox,
)

from parsagon.gui.controller import GUIController
from parsagon.gui.menu import MenuManager
from parsagon.settings import get_graphic

message_padding_constant = 0
message_width_proportion = 0.85
is_windows = os.name == "nt"
font_size_messages = 13 if not is_windows else 10
font_size_input = 15 if not is_windows else 11


try:
    from ctypes import windll

    app_id = "parsagon.parsagon.gui.1"
    windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)
except ImportError:
    pass


class CustomTextEdit(QTextEdit):
    def __init__(self, on_user_input_callback, parent=None):
        super().__init__(parent)
        self.on_user_input_callback = on_user_input_callback
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key.Key_Return:
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                super().keyPressEvent(event)
            else:
                self.on_user_input_callback()  # Call the passed function
        else:
            super().keyPressEvent(event)


class GUIWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.condition = Condition()

        self.setWindowTitle("Parsagon")
        self.setGeometry(100, 100, 400, 700)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)

        palette = self.palette()
        palette.setColor(self.backgroundRole(), Qt.GlobalColor.white)
        self.setPalette(palette)

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 12)
        main_layout.setSpacing(10)

        self.scroll_area = QScrollArea(self)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.message_container = QWidget()
        self.messages_layout = QVBoxLayout(self.message_container)
        self.messages_layout.setContentsMargins(0, 10, 0, 10)
        self.messages_layout.setSpacing(0)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(self.message_container)

        # Add a vertical spacer to push messages to the top
        self.spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.messages_layout.addSpacerItem(self.spacer)
        main_layout.addWidget(self.scroll_area)

        # Progress
        self.progress_container = QWidget()
        progress_layout = QVBoxLayout()
        self.progress_container.setLayout(progress_layout)
        self.progress_title = QLabel("Loading...", self)
        progress_layout.addWidget(self.progress_title)
        self.progress_bar = QProgressBar(self, minimum=0, maximum=100)
        progress_layout.addWidget(self.progress_bar)
        main_layout.addWidget(self.progress_container)
        self.progress_container.setVisible(False)

        input_wrapper_layout = QHBoxLayout()
        input_wrapper_layout.setContentsMargins(10, 0, 10, 0)

        input_border = QWidget()
        input_border.setContentsMargins(10, 2, 7, 2)
        input_border.setObjectName("inputBorder")
        input_border.setStyleSheet(
            """
            #inputBorder {
                border: 1px solid #CBCED2;
                border-radius: 14px;
            }
            QWidget {
                border: none;
            }
        """
        )
        input_wrapper_layout.addWidget(input_border)
        input_layout = QHBoxLayout(input_border)
        input_layout.setContentsMargins(0, 0, 0, 0)

        self.user_input_edit = CustomTextEdit(self.on_user_input, self)
        self.user_input_edit.setFont(QFont("Menlo", font_size_input))
        self.user_input_edit.setFixedHeight(43)
        input_layout.addWidget(self.user_input_edit, alignment=Qt.AlignmentFlag.AlignVCenter)
        self.user_input_edit.setEnabled(True)
        self.user_input_edit.setVisible(True)
        self.user_input_edit.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        pixmap = QPixmap(get_graphic("send@2x.png"))
        pixmap.setDevicePixelRatio(2.0)
        send_icon = QIcon(pixmap)
        self.send_button = QPushButton("", self)
        self.send_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.send_button.setIcon(send_icon)
        self.send_button.setIconSize(QSize(33, 33))
        self.send_button.clicked.connect(self.on_user_input)
        input_layout.addWidget(self.send_button, alignment=Qt.AlignmentFlag.AlignVCenter)

        self.loading_spinner = QLabel(self)
        self.loading_spinner.setAlignment(Qt.AlignmentFlag.AlignCenter)

        spinner_movie = QMovie(get_graphic("loading.gif"))
        spinner_movie.setScaledSize(QSize(33, 33))
        spinner_movie.currentPixmap().setDevicePixelRatio(2.0)
        self.loading_spinner.setMovie(spinner_movie)
        self.loading_spinner.setFixedHeight(33)
        spinner_movie.start()
        self.loading_spinner.setVisible(False)  # Initially hidden
        input_layout.addWidget(self.loading_spinner, alignment=Qt.AlignmentFlag.AlignVCenter)

        main_layout.addLayout(input_wrapper_layout)
        self.central_widget = QWidget()
        self.central_widget.setLayout(main_layout)
        self.setCentralWidget(self.central_widget)

        self.menu_manager = MenuManager(self)
        self.menu_manager.make_menu()

        # Signals
        self.controller = GUIController(self.condition)
        self.controller.set_loading_signal.connect(self.set_loading)
        self.controller.show_progress_signal.connect(self.show_progress)
        self.controller.set_progress_title_signal.connect(self.set_progress_title)
        self.controller.set_progress_signal.connect(self.set_progress)
        self.controller.update_text_signal.connect(self.update_text_ui)
        self.controller.request_input_signal.connect(self.focus_input)
        self.show()

        # Start the assist function after a delay to prevent layout blip upon opening
        self.timer = QTimer.singleShot(500, self.delayed_setup)

    def delayed_setup(self):
        self.controller.start()

    def on_user_input(self):
        user_input = self.user_input_edit.toPlainText()
        if user_input:
            self.update_text_ui(user_input, "user")
        with self.condition:
            self.controller.current_input = user_input
            self.user_input_edit.clear()
            self.condition.notify()

    @pyqtSlot(bool)
    def set_loading(self, loading):
        self.send_button.setVisible(not loading)
        self.loading_spinner.setVisible(loading)

    @pyqtSlot(bool)
    def show_progress(self, show_progress):
        self.progress_container.setVisible(show_progress)

    @pyqtSlot(str)
    def set_progress_title(self, title):
        self.progress_title.setText(title)

    @pyqtSlot(int)
    def set_progress(self, progress):
        self.progress_bar.setValue(progress)

    @pyqtSlot()
    def focus_input(self):
        self.user_input_edit.setFocus()

    @pyqtSlot(str, str, str)
    def update_text_ui(self, text, text_color="user", background_color="#33B1FF"):
        text = text.replace("\n", "<br>")
        is_user = text_color == "user"
        if is_user:
            text_color = "white"
        align_right = is_user

        message_edit = QTextEdit(self)
        message_edit.setFont(QFont("Menlo", font_size_messages))
        message_edit.insertHtml(text)
        message_edit.setReadOnly(True)
        message_edit.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        message_edit.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        message_edit.setWordWrapMode(QTextOption.WrapMode.WordWrap)
        message_edit.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        message_edit.setObjectName("message")
        message_edit.setStyleSheet(
            f"""
            #message {{
                border: none;
                background-color: transparent;
                color: {text_color}
            }}
        """
        )

        outer_callout_layout = QVBoxLayout()
        outer_callout_layout.setContentsMargins(0, 0, 0, 0)
        outer_callout_layout.setSpacing(0)

        callout = QFrame(self)
        callout.setObjectName("messageContainer")
        callout.setStyleSheet(
            f"""
            #messageContainer {{
                border-radius: 12px;
                background-color: {background_color};
            }}
        """
        )
        callout_tail = QLabel(self)
        path_suffix = "_user" if align_right else ""
        callout_arrow_path = get_graphic(f"callout_arrow{path_suffix}@2x.png")
        pixmap = QPixmap(callout_arrow_path)
        pixmap.setDevicePixelRatio(2.0)
        callout_tail.setPixmap(pixmap)
        callout_tail_width = pixmap.width() // 2
        padding_amount = 7
        callout_tail.setFixedSize(QSize(callout_tail_width + padding_amount, pixmap.height()))
        callout_tail.setObjectName("calloutTail")
        callout_tail.setAlignment(Qt.AlignmentFlag.AlignLeft if not align_right else Qt.AlignmentFlag.AlignRight)
        padding_side = "right" if align_right else "left"
        callout_tail.setStyleSheet(
            f"""
            #calloutTail {{
                padding-{padding_side}: {padding_amount}px;
            }}
        """
        )

        inner_callout_layout = QVBoxLayout(callout)
        inner_callout_layout.setContentsMargins(5, 5, 5, 5)
        inner_callout_layout.setSpacing(0)
        inner_callout_layout.addWidget(message_edit)

        outer_callout_layout.addWidget(callout)
        outer_callout_layout.addWidget(
            callout_tail, alignment=Qt.AlignmentFlag.AlignLeft if not align_right else Qt.AlignmentFlag.AlignRight
        )

        hbox = QHBoxLayout()
        hbox.setContentsMargins(10, 0, 10, 0)
        if align_right:
            hbox.addSpacerItem(QSpacerItem(35, 20, QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum))
            hbox.addLayout(outer_callout_layout)
        else:
            hbox.addLayout(outer_callout_layout)
            hbox.addSpacerItem(QSpacerItem(35, 20, QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum))

        self.messages_layout.addLayout(hbox)
        message_edit.show()
        inner_callout_layout.update()
        callout.show()
        message_edit.setFixedHeight(int(message_edit.document().size().height()) + message_padding_constant)

        # Update the spacer
        self.messages_layout.removeItem(self.spacer)
        self.messages_layout.addSpacerItem(self.spacer)

        self.scroll_area.verticalScrollBar().setValue(self.scroll_area.verticalScrollBar().maximum())



import functools
import json
import re

from httpx import codes


class MockResponse:
    def __init__(self, status_code, json_body=None, text_body=None):
        self.status_code = status_code
        self.json_body = json_body
        self.text_body = text_body
        assert not (json_body and text_body)

    def json(self):
        if not self.json_body:
            raise json.JSONDecodeError("No JSON body", "", 0)
        return self.json_body

    @property
    def is_success(self):
        return codes.is_success(self.status_code)

    @property
    def text(self):
        return json.dumps(self.json_body) if self.json_body else self.text_body or ""


not_found_pipeline_name = "my_incorrect_name"
not_found_pipeline_id = 44
not_found_response = MockResponse(404, json_body={"detail": "Not found."})


def mock_httpx_method_func(*args, **kwargs):
    """
    A mock that is used for httpx post, get, put, patch, and delete methods.  The "method" kwarg differentiates which mock is being called, and the "mock_options" kwarg provides further options pertaining to the desired behavior of the mock. The following keys are supported:

    - code_to_return: For operations that require code to be returned, this will be used.
    """
    method = kwargs["method"].lower()
    mock_options = kwargs["mock_options"]
    url = args[0]

    if match := re.search(r"/pipelines/name/(.+)/code/$", url):
        assert method == "post"
        pipeline_name = match.group(1)
        if pipeline_name == not_found_pipeline_name:
            return not_found_response
        return MockResponse(
            200,
            json_body={
                "code": mock_options["code_to_return"],
            },
        )

    if match := re.search(r"/pipelines/name/(.+)/", url):
        assert method == "get"
        pipeline_name = match.group(1)
        if pipeline_name == not_found_pipeline_name:
            return not_found_response
        return MockResponse(
            200,
            json_body={
                "id": hash(pipeline_name) % 1000,
                "name": pipeline_name,
            },
        )

    if match := re.search(r"/pipelines/(\d+)/$", url):
        assert method == "delete"
        pipeline_id = match.group(1)
        if pipeline_id == not_found_pipeline_id:
            return not_found_response
        return MockResponse(204)

    raise Exception("Unknown combination of method and url: %s %s" % (method, url))


def install_api_mocks(mocker, mock_options=None):
    """
    Installs mocks for the backend. The "mock_options" kwarg can be used to customize responses. See mock_httpx_method_func for details.
    """

    if mock_options is None:
        mock_options = {}

    # NOTE: Remember to update this if we add code that introduces other calling points to the backend, or if we change the way these httpx method functions are imported.
    for method in ["get", "post", "put", "patch", "delete"]:
        mocker.patch(
            "parsagon.api.httpx.%s" % method,
            functools.partial(mock_httpx_method_func, method=method, mock_options=mock_options),
        )



import argparse
import logging

import pytest

from parsagon import delete, run
from parsagon.main import main
from parsagon.tests.api_mocks import install_api_mocks, not_found_pipeline_name


def test_pipeline_delete(mocker):
    install_api_mocks(mocker)
    delete("test_program")


def test_pipeline_not_found(mocker, debug_logs):
    install_api_mocks(mocker, {"code_to_return": 'raise Exception("Should not exec this code if pipeline not found.")'})

    # On delete
    main(["delete", not_found_pipeline_name, "-y"])
    assert f"A program with name {not_found_pipeline_name} does not exist." in debug_logs.text
    debug_logs.clear()

    # On attempted run
    main(["run", not_found_pipeline_name])
    assert f"A program with name {not_found_pipeline_name} does not exist." in debug_logs.text



from parsagon.api import get_program_sketches, get_pipeline, create_pipeline, create_custom_function, delete_pipeline
from parsagon.exceptions import APIException, ProgramNotFoundException
from parsagon.executor import Executor, custom_functions_to_descriptions
from parsagon.print import assistant_print, ask, input, confirm
from parsagon.secrets import extract_secrets


def create_program(task, headless=False, undetected=False, program_name=None, assume_yes=False):
    if program_name:
        try:
            get_pipeline(program_name)
            assistant_print(f"A program with the name {program_name} already exists.")
            return {
                "success": False,
                "outcome": f"A program with the name {program_name} already exists. The user must choose a different name.",
            }
        except ProgramNotFoundException:
            pass
    assistant_print("Creating a program based on your specifications...")
    task, secrets = extract_secrets(task)
    program_sketches = get_program_sketches(task)

    full_program = program_sketches["full"]
    abridged_program = program_sketches["abridged"]
    pseudocode = program_sketches["pseudocode"]
    assistant_print(f"Here's an outline of what the program does:\n\nSummary: {task}\n\nSteps:\n\n{pseudocode}\n\n")
    if assume_yes:
        infer = True
    else:
        approval = confirm("Confirm the program does what you want")
        if not approval:
            feedback = ask("What do you want the program to do differently?")
            return {"success": False, "outcome": "User canceled program creation", "user_feedback": feedback}
        infer_response = ask(
            "To complete the program, Parsagon must visit the page(s) mentioned above and identify the exact web elements to interact with. Hit ENTER to let Parsagon do this on its own, or type MANUAL to show Parsagon the relevant elements by clicking on them",
            choices=["", "MANUAL"],
            show_choices=False,
        )
        infer = infer_response != "MANUAL"

    assistant_print(f"Now executing the program to identify web elements to be scraped:")
    args = ", ".join(f"{k}={repr(v)}" for k, v in secrets.items())
    abridged_program += f"\n\noutput = func({args})\n"  # Make the program runnable

    # Execute the abridged program to gather examples
    executor = Executor(task, headless=headless, infer=infer, use_uc=undetected)
    example_output = executor.execute(abridged_program)

    # The user must select a name
    while True:
        if not program_name:
            program_name = ask("Name this program to save, or press enter without typing a name to DISCARD")
        if program_name:
            assistant_print(f"Saving program as {program_name}")
            try:
                pipeline = create_pipeline(program_name, task, full_program, pseudocode, secrets)
            except APIException as e:
                if isinstance(e.value, list) and "Pipeline with name already exists" in e.value:
                    assistant_print("A program with this name already exists. Please choose another name.")
                    program_name = None
                    continue
                else:
                    raise e
            pipeline_id = pipeline["id"]
            try:
                for call_id, custom_function in executor.custom_functions.items():
                    description = custom_functions_to_descriptions.get(custom_function.name)
                    description = " to " + description if description else ""
                    assistant_print(f"  Saving function{description}...")
                    create_custom_function(pipeline_id, call_id, custom_function)
                assistant_print(f"Saved.")
                break
            except:
                delete_pipeline(pipeline_id)
                assistant_print(f"An error occurred while saving the program. The program has been discarded.")
                return {"success": False, "outcome": f"User decided not to save the program"}
        else:
            assistant_print("Discarded program.")
            return {"success": False, "outcome": f"User decided not to save the program"}

    assistant_print("Done.")
    return {
        "success": True,
        "outcome": f"Program successfully saved with name {program_name}",
        "program_name": program_name,
        "example_output": example_output,
    }



import multiprocessing
import os

# Pyinstaller fix - must come as early as possible, even above other imports.
multiprocessing.freeze_support()

# Set environment variables
import gui_env  # noqa

assert os.environ["GUI_ENABLED"] == "1", "GUI_ENABLED must be set to 1 in gui_env.py"
import sys

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication

from parsagon.gui import GUIWindow
from parsagon.settings import get_graphic

is_windows = os.name == "nt"

if __name__ == "__main__":
    app = QApplication(sys.argv)
    if is_windows:
        app.setWindowIcon(QIcon(get_graphic("win_icon.ico")))
    gui_window = GUIWindow()
    sys.exit(app.exec())



import os
import webbrowser
from datetime import datetime

from PyQt6.QtGui import QAction, QKeySequence
from PyQt6.QtWidgets import QApplication, QTextEdit, QMessageBox


docs_url = "https://parsagon.io/docs/pipelines/overview"
is_windows = os.name == "nt"


class MenuManager:
    def __init__(self, window):
        self.window = window

    def make_menu(self):
        window = self.window
        menu_bar = window.menuBar()

        # APP NAME / FILE
        app_menu = menu_bar.addMenu("File" if is_windows else "&Parsagon")

        about_action = QAction("About", window)
        app_menu.addAction(about_action)
        about_action.triggered.connect(self.show_about_dialog)

        # EDIT
        edit_menu = menu_bar.addMenu("&Edit")

        # Undo action
        undo_action = QAction("&Undo", window, shortcut=QKeySequence.StandardKey.Undo)
        undo_action.triggered.connect(self.undo)
        edit_menu.addAction(undo_action)

        # Redo action
        redo_action = QAction("&Redo", window, shortcut=QKeySequence.StandardKey.Redo)
        redo_action.triggered.connect(self.redo)
        edit_menu.addAction(redo_action)

        edit_menu.addSeparator()

        # Cut action
        cut_action = QAction("Cu&t", window, shortcut=QKeySequence.StandardKey.Cut)
        cut_action.triggered.connect(self.cut_text)
        edit_menu.addAction(cut_action)

        # Copy action
        copy_action = QAction("&Copy", window, shortcut=QKeySequence.StandardKey.Copy)
        copy_action.triggered.connect(self.copy_text)
        edit_menu.addAction(copy_action)

        # Paste action
        paste_action = QAction("&Paste", window, shortcut=QKeySequence.StandardKey.Paste)
        paste_action.triggered.connect(self.paste_text)
        edit_menu.addAction(paste_action)

        # Select all action
        select_all_action = QAction("Select &All", window, shortcut=QKeySequence.StandardKey.SelectAll)
        select_all_action.triggered.connect(self.select_all_text)
        edit_menu.addAction(select_all_action)

        # HELP
        help_menu = menu_bar.addMenu("&Help")

        documentation_action = QAction("Help", window)
        documentation_action.triggered.connect(self.show_documentation)
        help_menu.addAction(documentation_action)

    def cut_text(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.cut()

    def copy_text(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.copy()

    def paste_text(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.paste()

    def undo(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.undo()

    def redo(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.redo()

    def select_all_text(self):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QTextEdit):
            focused_widget.selectAll()

    def show_about_dialog(self):
        version = os.environ.get("VERSION", "unknown")
        year = datetime.now().year
        message = f"""
        Version {version}<br>
        Copyright © {year} Parsagon, Inc.
        """

        # Title appears to be broken
        box = QMessageBox()
        box.about(self.window, "", message)
        box.setWindowTitle("About Parsagon")

    def show_documentation(self):
        webbrowser.open(docs_url)



from parsagon.api import (
    edit_program_sketch,
    get_pipeline,
    get_pipeline_code_parts,
    update_pipeline,
    add_examples_to_custom_function,
)
from parsagon.executor import Executor, custom_functions_to_descriptions
from parsagon.print import assistant_print
from rich.prompt import Confirm, Prompt


def edit_program(task, program_name, variables={}, verbose=False):
    pipeline = get_pipeline(program_name)
    pipeline_code_parts = get_pipeline_code_parts(pipeline["id"])
    custom_functions = {
        custom_function["call_id"]: custom_function["code"]
        for custom_function in pipeline_code_parts["custom_functions"]
    }
    program_sketches = edit_program_sketch(pipeline_code_parts["program_sketch"], task)
    full_program = program_sketches["full"]
    abridged_program = program_sketches["abridged"]
    pseudocode = program_sketches["pseudocode"]
    assistant_print(f"Edited the program to do the following:\n\n{pseudocode}\n")
    approval = Confirm.ask("Confirm the program does what you want")
    if not approval:
        feedback = Prompt.ask("What do you want the program to do differently?")
        return {"success": False, "outcome": "User canceled program editing", "user_feedback": feedback}

    assistant_print(f"Now executing the program to identify web elements to be scraped:")
    args = ", ".join(f"{k}={repr(v)}" for k, v in variables.items())
    abridged_program += f"\n\noutput = func({args})\n"  # Make the program runnable

    # Execute the abridged program to gather examples
    executor = Executor(task, function_bank=custom_functions)
    executor.execute(abridged_program)

    program_name_input = Prompt.ask(
        f'Type "{program_name}" to update this program, or press enter without typing a name to CANCEL',
        choices=[program_name, ""],
    )
    if not program_name_input:
        assistant_print("Discarded edits.")
        return {"success": False, "outcome": f"User decided not to save the edits"}

    pipeline_id = pipeline["id"]
    try:
        for call_id, custom_function in executor.custom_functions.items():
            debug_suffix = f" ({custom_function.name})"
            description = custom_functions_to_descriptions.get(custom_function.name)
            description = " to " + description if description else ""
            if verbose:
                description += debug_suffix
            assistant_print(f"  Saving function{description}...")
            add_examples_to_custom_function(pipeline_id, call_id, custom_function, True)
        assistant_print(f"Finalizing program...")
        update_pipeline(
            pipeline_id, {"program_sketch": full_program, "abridged_sketch": abridged_program, "pseudocode": pseudocode}
        )
        assistant_print(f"Saved.")
    except Exception as e:
        error_message = f"An error occurred while saving the program. The program was not updated. {repr(e)}"
        assistant_print(error_message)
        return {"success": False, "outcome": error_message}

    return {"success": True, "outcome": f"Program successfully edited", "program_name": program_name}



import logging

import pytest


@pytest.fixture
def debug_logs(caplog, mocker):
    """
    The caplog fixture does not work with Parsagon's logging configuration - override the logging configuration, set sensitivity to debug, and return the caplog fixture.
    """
    mocker.patch("parsagon.main.configure_logging", lambda level: logging.basicConfig(level=logging.DEBUG))
    caplog.set_level(logging.DEBUG)
    return caplog
